import itertools
import numpy as np
from numpy import ix_
import math
import tempfile
from dask import delayed
from dask.distributed import as_completed
from sklearn.externals import joblib

from . import bilp
from . import common

import functools
print = functools.partial(print, flush=True)

def reindex(orig_inxs, order_inxs):
    return list(np.array(orig_inxs)[order_inxs])

def find_P(path_D, o, t, bilp_test=False, prune_history=None, check_and_recurse=True, bilp_method="mos2", client=None):
    def load_D():
        return np.genfromtxt(path_D, delimiter=",")

    bilp_res = bilp.bilp(load_D(),method=bilp_method)
    k_optimal = bilp_res[0]
    print("k",k_optimal)

    def _exhaustive_search(left_perm, search_perm, right_perm):
        all_search_perms = [list(perm) for perm in itertools.permutations(
            search_perm, len(search_perm))]

        def local_calc_k(local_search_perm): return _calc_k(
            left_perm + local_search_perm + right_perm)
        results = list(map(local_calc_k, all_search_perms))
        k_values = []
        for k_value, Dperm_right in results:
            k_values.append(k_value)
        k = np.min(k_values)
        P = [left_perm + opt_perm + right_perm for opt_perm in np.array(
            all_search_perms)[np.where(k_values == k)[0]].tolist()]
        return k, P

    def _calc_k(left_perm, right_perm=[]):
        D = load_D()
        perm = left_perm + right_perm
        Dperm = D[ix_(perm, perm)]
        # Left perm is optimal, but right perm is not, so we'll set it to the perfect submatrix.
        # This could be replaced to a call to bilp for a tighter bound
        if len(right_perm) > 0:
            rn = len(right_perm)
            right_inxs = np.arange(len(left_perm),Dperm.shape[0])
            Dperm_right = Dperm[ix_(right_inxs, right_inxs)]
            Dperm[ix_(right_inxs, right_inxs)] = np.triu(np.ones((rn, rn)), 1)
        else:
            Dperm_right = None
        Dperfect = np.triu(np.ones((D.shape[0], D.shape[0])), 1)
        k = int(np.sum(np.abs(Dperfect-Dperm)))
        return k, Dperm_right
    
    def _check(left_perm,add_left_perm,prev_right_perm):
        if type(add_left_perm) == tuple:
            add_left_perm = list(add_left_perm)
        
        # move the contents of add_left_perm out of prev_right_perm, this wrinkle is done to
        # save memory
        right_perm = list(set(prev_right_perm) - set(add_left_perm))
        
        perm_k, Dperm_right = _calc_k(left_perm+add_left_perm, right_perm=right_perm)
        if perm_k <= k_optimal:
            if bilp_test: # extra bilp test
                kbilp_res = bilp.bilp(Dperm_right,method=bilp_method)
                kbilp = kbilp_res[0]
                if kbilp+perm_k <= k_optimal: # still need to continue
                    return True, 0, add_left_perm, right_perm
            else:
                return True, 0, left_perm, add_left_perm, right_perm
            
        # pruned
        skipped = math.factorial(len(right_perm))
        return False, skipped, left_perm, add_left_perm, right_perm
            
    
    def _get_futures(left_perm,right_perm):
        def _multi_check(add_left_perms):
            results = []
            for add_left_perm in add_left_perms:
                results.append(_check(left_perm,add_left_perm,right_perm))
            return results
        
        add_left_perms = list(itertools.permutations(right_perm, o))
        num_add_left_perms = common.nPr(len(right_perm),o)
        desired_number_chunks = 1000
        chunk_length = round(num_add_left_perms*1./desired_number_chunks)
        if chunk_length == 0:
            chunk_length = len(add_left_perms)
        add_left_perms_chunks = common.chunks(add_left_perms,chunk_length)
        futures = client.map(_multi_check,add_left_perms_chunks)
        return futures
    
    def _iterative_find_P():
        D = load_D()
        right_perm = list(range(D.shape[0]))
        P = []
        skipped = 0
        futures = _get_futures([],right_perm)
        prune_history_record = {}
        seq = as_completed(futures)
        for future in seq:
            print(seq.count())
            for result_check, result_skipped, result_left_perm, result_add_left_perm, result_right_perm in future.result():
                if result_check:
                    if len(result_right_perm) <= o or math.factorial(len(result_right_perm)) <= t:
                        k, Pnew = _exhaustive_search(result_left_perm + result_add_left_perm, result_right_perm, [])
                        if k == k_optimal:
                            P.extend(Pnew)
                        elif k < k_optimal:
                            print(k,k_optimal,Pnew)
                            raise Exception(
                                "Error! The k value that you found was less than the optimal, which should never happen.")
                    else:
                        new_futures = _get_futures(result_left_perm + result_add_left_perm,result_right_perm)
                        for new_future in new_futures:
                            seq.add(new_future)
                #else:
                #    print('pruned')
                #    skipped += result_skipped
                #    if prune_history:
                #        prune_history_record["skipped"] = skipped
                #        prune_history_record[tuple(result_left_perm+result_add_left_perm)] = True
                #        joblib.dump(prune_history_record,"/dev/shm/prune_history_record.joblib.z")

        return P, skipped
    
    P, skipped = _iterative_find_P()
    return k_optimal, P, skipped



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mar 25

@author: yoichi_izunaga + amy langville + paul anderson
"""

import numpy as np
from gurobipy import *
import itertools

from .common import *

def optimize_weights(M,w_values,min_support,max_solutions=100000):
    k_values = {}
    P_values = {}
    p_values = {}
    D_values = {}
    completed = {}
    nunique = 0
    nskipped = 0
    for w in itertools.combinations_with_replacement(w_values,M.shape[1]):
        w = np.array(w)*1./sum(w)
        w_tuple = tuple(w)
        D = define_D(M,w,min_support)
        D_tuple = tuple(map(tuple, D))
        if D_tuple in completed:
            eq_w = completed[D_tuple]
            k_values[w_tuple] = k_values[eq_w]
            p_values[w_tuple] = p_values[eq_w]
            P_values[w_tuple] = P_values[eq_w]
            nskipped += 1
            continue
        k,p,P = bilp(D,max_solutions=max_solutions)

        k_values[w_tuple] = k
        p_values[w_tuple] = p
        P_values[w_tuple] = P
        D_values[w_tuple] = D

        completed[D_tuple] = w_tuple

        nunique += 1
    
    # Find the best
    best_k = np.Inf
    best_w = None
    for w, k in k_values.items():
        if k < best_k:
            best_k = k
            best_w = w
    
    return best_w, k_values, P_values, p_values, D_values

def bilp_from_data(M,w,min_support,max_solutions=1):
    D = define_D(M,w,min_support)
    return bilp(D,max_solutions=max_solutions)

def bilp(D,max_solutions=None,smartk=None,method=["orig","mos2","ref"][0]):
    if method == "mos2":
        return bilp_mos2(D,max_solutions=max_solutions,smartk=smartk)
    elif method == "orig":
        return bilp_orig(D,max_solutions=max_solutions,smartk=smartk)
    elif method == "ref":
        return bilp_ref(D,max_solutions=max_solutions,smartk=smartk)
    else:
        raise Exception("Invalid method "+str(method))

def bilp_orig(D,max_solutions=None,smartk=None):
    n = len(D[0])
    m = int(np.sum(D))

    AP = Model("rankability")
    
    AP.setParam( 'OutputFlag', False )

    x = {}
    y = {}
    for i in range(n):
        for j in range(n):
            x[i,j] = AP.addVar(lb=0,vtype=GRB.BINARY,ub=1,name="x(%s,%s)"%(i,j)) #binary
            y[i,j] = AP.addVar(lb=0,vtype=GRB.BINARY,ub=1,name="y(%s,%s)"%(i,j))  #binary
    AP.update()

    for i in range(n):
        for j in range(n):
            AP.addConstr(x[i,j] + y[i,j] <= 1)
            AP.addConstr(x[i,j] <= 1 - D[i,j])
            AP.addConstr(y[i,j] <= D[i,j])
    for j in range(1,n):
        for i in range(0,j):
            AP.addConstr(D[i,j] + x[i,j] - y[i,j] + D[j,i] + x[j,i] - y[j,i] == 1)

    for i in range(n):
        for j in range(n):
            for k in range(n):
                if j != i and k != j and k != i:
                    AP.addConstr(D[i,j] + x[i,j] - y[i,j] + D[j,k] + x[j,k] - y[j,k] + D[k,i] + x[k,i] - y[k,i] <= 2)

    if smartk == None:
        smartk = (n*n-n)/2
    AP.addConstr(quicksum(x[i,j] for i in range(n) for j in range(n)) + quicksum(y[i,j] for i in range(n) for j in range(n)) <= smartk)

    AP.update()

    AP.setObjective(quicksum(x[i,j] for i in range(n) for j in range(n)) + quicksum(y[i,j] for i in range(n) for j in range(n)),GRB.MINIMIZE)
    AP.update()
    
    get_sol_x_func = get_sol_x_by_x(x,n)
    get_sol_y_func = get_sol_y_by_y(y,n)
       
    return _run(D,AP,get_sol_x_func,get_sol_y_func,max_solutions)

def _run(D,model,get_sol_x_func,get_sol_y_func,max_solutions,obj2k_func=lambda obj: int(obj)):
    # Limit how many solutions to collect
    if max_solutions != None:
        model.setParam(GRB.Param.PoolSolutions, max_solutions)
        # Limit the search space by setting a gap for the worst possible solution that will be accepted
        model.setParam(GRB.Param.PoolGap, .5)
        # do a systematic search for the k-best solutions
        if max_solutions != None:
            model.setParam(GRB.Param.PoolSearchMode, 2)
        model.update()

    model.optimize()

    k=obj2k_func(model.objVal)

    if max_solutions == None:
        return k,[]
    
    P = _get_P(D,model,get_sol_x_func,get_sol_y_func)
    return k, P

def round_Xn(Xn,mult=10):
    return round(mult*Xn)*1./mult

def get_sol_x_by_x(x,n):
    return lambda: np.reshape([round_Xn(x[i,j].Xn) for i in range(n) for j in range(n)],(n,n))

def get_sol_y_by_y(y,n):
    return lambda: np.reshape([round_Xn(y[i,j].Xn) for i in range(n) for j in range(n)],(n,n))

def get_sol_x_by_z(D,z,n):
    return lambda: np.reshape([1 if round_Xn(z[i,j].Xn) == 1 and D[i,j] == 0 and i != j else 0 for i in range(n) for j in range(n)],(n,n))

def get_sol_y_by_z(D,z,n):
    return lambda: np.reshape([1 if round_Xn(z[i,j].Xn) == 0 and D[i,j] == 1 and i != j else 0 for i in range(n) for j in range(n)],(n,n))

def get_sol_z_by_z(D,z,n):
    return lambda: np.reshape([round_Xn(z[i,j].Xn) for i in range(n) for j in range(n)],(n,n))

"""If zij = 0 and dij = 1, then set yij = 1. If zij = 1 and dij = 0,
292 then set xij = 1. Then k is the number of nonzeros in X plus the number of nonzeros
293 in Y, i.e., k = nnz(X) + nnz(Y)."""
    
def _get_P(D,model,get_sol_x_func,get_sol_y_func):
    # Print number of solutions stored
    nSolutions = model.SolCount

    # Print objective values of solutions and create a list of those that have the same objective value as the optimal
    alternative_solutions = []
    for e in range(nSolutions):
        model.setParam(GRB.Param.SolutionNumber, e)
        if model.PoolObjVal == model.objVal:
            alternative_solutions.append(e)
    
    # print all alternative solutions
    P = {}
    for e in alternative_solutions:
        model.setParam(GRB.Param.SolutionNumber, e)
        sol_x = get_sol_x_func()
        sol_y = get_sol_y_func()
        rowsum=np.sum(D+sol_x-sol_y,axis=0)
        ranking=np.argsort(rowsum)+1
        key = tuple([int(item) for item in ranking])
        if key not in P:
            P[key] = True
    P = [list(perm) for perm in P.keys()]
    return P

def bilp_mos2(A,max_solutions=None,smartk=None):
    n = len(A[0])
    m = int(np.sum(A))
    AP = Model("rankabilityAltFormReferee")
    AP.setParam( 'OutputFlag', False )
    z = {}
    for i in range(n):
        for j in range(n):
            z[i,j] = AP.addVar(lb=0,vtype=GRB.BINARY,ub=1,name="z(%s,%s)"%(i,j)) #binary
    AP.update()

    for j in range(1,n):
        for i in range(0,j):
            AP.addConstr(z[i,j] + z[j,i] == 1)

    for i in range(n):
        for j in range(n):
            for k in range(n):
                if j != i and k != j and k != i:
                    AP.addConstr(z[i,j] + z[j,k] + z[k,i] <= 2)

    AP.update()

    AP.setObjective(quicksum(A[i,j]*z[i,j] for i in range(n) for j in range(n)),GRB.MAXIMIZE)
    AP.update()
               
    get_sol_x_func = get_sol_x_by_z(A,z,n)
    get_sol_y_func = get_sol_y_by_z(A,z,n)
    get_sol_z_func = get_sol_z_by_z(A,z,n)
    
    obj2k_func=lambda obj: int(np.count_nonzero(get_sol_x_func())) + int(np.count_nonzero(get_sol_y_func()))
       
    results = _run(A,AP,get_sol_x_func,get_sol_y_func,max_solutions,obj2k_func=obj2k_func)
    #print("D:")
    #print(A)
    #print("z:")
    #print(get_sol_z_func())
    #print("x:")
    #print(get_sol_x_func())
    #print("y:")
    #print(get_sol_y_func())
    
    return results
               
def bilp_ref(A,max_solutions=None,smartk=None):
    n = len(A[0])
    m = int(np.sum(A))

    AP = Model("rankabilityAltFormReferee")
    AP.setParam( 'OutputFlag', False )
    z = {}
    for i in range(n):
        for j in range(n):
            z[i,j] = AP.addVar(lb=0,vtype=GRB.BINARY,ub=1,name="z(%s,%s)"%(i,j)) #binary
    AP.update()

    for j in range(1,n):
        for i in range(0,j):
            AP.addConstr(z[i,j] + z[j,i] == 1)

    for i in range(n):
        for j in range(n):
            for k in range(n):
                if j != i and k != j and k != i:
                    AP.addConstr(z[i,j] + z[j,k] + z[k,i] <= 2)

    AP.update()

    AP.setObjective(quicksum(A[i,j]*z[i,j] for i in range(n) for j in range(n)),GRB.MAXIMIZE)
    AP.update()
    
    get_sol_x_func = get_sol_x_by_z(A,z,n)
    get_sol_y_func = get_sol_y_by_z(A,z,n)
    
    obj2k_func=lambda obj: np.count_nonzero(get_sol_x_func()) + np.count_nonzero(get_sol_y_func())
       
    return _run(A,AP,get_sol_x_func,get_sol_y_func,max_solutions,obj2k_func=obj2k_func)
               




import math
import numpy as np
import json
import sys
import string
import random

np.set_printoptions(threshold=sys.maxsize)

def random_generator(size=6, chars=string.ascii_uppercase + string.digits):
    return ''.join(random.choice(chars) for x in range(size))

def nCr(n,r):
    f = math.factorial
    return f(n) // f(r) // f(n-r)
    #return comb(n,r)
    
def nPr(n,r):
    f = math.factorial
    return f(n) // f(n-r)

def permute_D(D,perm):
    return D[perm,:][:,perm]

def calc_k(D,max_value=None):
    if not max_value:
        max_value = np.max(D)
        if max_value == 0:
            max_value = 1
    perfectRG=np.triu(max_value*np.ones((D.shape[0],D.shape[0])),1).astype(int)
    k = np.sum(np.abs(perfectRG-D))
    return k

def define_D(M,w,min_support):
    n = M.shape[0]
    D = np.zeros((n,n))
    for i in range(n):
        for j in range(n):
            if i == j:
                continue
            c = 0
            for k in range(len(w)):
                c += w[k]*1.0*int(M[i,k] > M[j,k])
            D[i,j] = int(c > min_support)
            
    return D

def len_chunks(l,n):
    return len(range(0,len(l),n))

def chunks(l, n):
    chunks_ = []
    for i in range(0, len(l), n):
        #yield l[i:i + n]
        chunks_.append(l[i:i + n])
    return chunks_

def chunks_generator(l,n):
    for i in range(0, len(l), n):
        yield l[i:i + n]
        
def as_json(k,P,other={},p=None):
    if len(P) > 0 and np.min(P) == 0:
        P = (np.array(P,dtype=int)+1).tolist()
    if p == None:
        p = len(P)
    return json.dumps({"k":k,"p": p, "P":P,"other":other})

def json_string(k,P,other={},p=None):
    p = len(P)
    if len(P) > 0 and np.min(P) == 0:
        P = (np.array(P,dtype=int)+1).tolist()
    if p == None:
        p = len(P)
    k = int(k)
    indent = "    "
    instance_as_string = "{\n"
    instance_as_string += indent + json.dumps({"k": k}).replace("{","").replace("}","")+",\n"
    instance_as_string += indent + json.dumps({"p": p}).replace("{","").replace("}","")+",\n"
    instance_as_string += indent + '"P": \n'
    P_as_string = np.array2string(np.array(P,dtype=int),separator=",",max_line_width=np.Inf).replace("[","[\n",1)
    lines = P_as_string.split("\n")
    for i,line in enumerate(lines):
        add_indent = indent+indent
        if i > 0:
            add_indent += indent
        instance_as_string += add_indent + line + "\n"
    instance_as_string = instance_as_string[:-2] + "],\n"
    instance_as_string += indent + json.dumps({"other": other})[1:-1]+"\n"
    instance_as_string += "}"
    return instance_as_string
    
        
class Search:
    def to_json(self):
        solution = {}
        solution["k"] = int(self.k)
        solution["p"] = len(self.P)
        solution["P"] = [[int(v) + 1 for v in perm] for perm in self.P]
        solution["other"] = self.other_to_dict()
        return solution
    
    # Override this if the search subclass does not keep track of thse things
    def other_to_dict(self):
        return {"skipped": self.skipped, "searched": self.searched, "exact_solution_found": self.exact_solution_found}
        



import itertools
import numpy as np
from numpy import ix_
import math
import tempfile
from dask import delayed
from dask.distributed import as_completed
from sklearn.externals import joblib
import copy
import uuid
import os
import math

from operator import itemgetter

from . import bilp
from . import common

import functools
print = functools.partial(print, flush=True)

def reindex(orig_inxs, order_inxs):
    return list(np.array(orig_inxs)[order_inxs])

def find_P(path_D, o, t, bilp_test=False, bilp_method="mos2", client=None, max_search_space=100000, prune_history=False, uuid1="", verbose=False):
    def load_D():
        return np.genfromtxt(path_D, delimiter=",")
    orig_right_perm = list(range(load_D().shape[0]))
    bilp_res = bilp.bilp(load_D(),method=bilp_method)
    k_optimal = bilp_res[0]
    if verbose:
        print("k",k_optimal)
    add_left_perms_grouped_file = "/dev/shm/add_left_perms_grouped"+uuid1+".joblib"
    add_left_perms_grouped_best_ks_file = "/dev/shm/add_left_perms_grouped_best_ks"+uuid1+".joblib"
    if verbose:
        print(add_left_perms_grouped_file,add_left_perms_grouped_best_ks_file)
        
    def _exhaustive_search(left_perm, search_perm, right_perm):
        all_search_perms = [list(perm) for perm in itertools.permutations(
            search_perm, len(search_perm))]

        def local_calc_k(local_search_perm): return _calc_k(
            left_perm + local_search_perm + right_perm)
        results = list(map(local_calc_k, all_search_perms))
        k_values = []
        for k_value, Dperm_right in results:
            k_values.append(k_value)
        k = np.min(k_values)
        optimal_inxs = np.where(k_values == k)[0]
        P = [left_perm + opt_perm + right_perm for opt_perm in np.array(
            all_search_perms)[optimal_inxs].tolist()]
        return k, P, [list(search_perm) for search_perm in np.array(all_search_perms)[optimal_inxs]]

    def _calc_k(left_perm, right_perm=[]):
        perm = left_perm + right_perm
        D = load_D()
        Dperm = D[ix_(perm, perm)]
        # Left perm is optimal, but right perm is not, so we'll set it to the perfect submatrix.
        # This could be replaced to a call to bilp for a tighter bound
        if len(right_perm) > 0:
            rn = len(right_perm)
            right_inxs = np.arange(len(left_perm),Dperm.shape[0])
            Dperm_right = Dperm[ix_(right_inxs, right_inxs)]
            Dperm[ix_(right_inxs, right_inxs)] = np.triu(np.ones((rn, rn)), 1)
        else:
            Dperm_right = None
        Dperfect = np.triu(np.ones((Dperm.shape[0], Dperm.shape[0])), 1)
        k = int(np.sum(np.abs(Dperfect-Dperm)))
        return k, Dperm_right
    
    def generate_right_perm(calc_k_left_perm):
        calc_k_right_perm = list(set(orig_right_perm) - set(calc_k_left_perm))
        _dict = {}
        for key in list(itertools.permutations(calc_k_right_perm, o)):
            _dict[tuple(sorted(key))] = True
        return list(_dict.keys())
    
    def _check(left_perm,add_left_perm,add_left_perms_grouped):
        def _convert(perm):
            actual_perm = []
            for key in perm:
                actual_perm += add_left_perms_grouped[key][0]
            return actual_perm  
    
        if type(add_left_perm) == tuple:
            add_left_perm = [add_left_perm]
        
        # move the contents of add_left_perm out of prev_right_perm, this wrinkle is done to
        # save memory
        if len(left_perm) == 0:
            calc_k_left_perm = _convert(add_left_perm)
        else:
            calc_k_left_perm = _convert(left_perm+add_left_perm)
        calc_k_right_perm = list(set(orig_right_perm) - set(calc_k_left_perm)) #list(set(prev_right_perm) - set(add_left_perm))
        #right_perm = []
        #partial_add_left_perm = set(add_left_perm[0])
        #for key in prev_right_perm:
        #    if partial_add_left_perm.isdisjoint(key):
        #        right_perm.append(key)
            
        perm_k, Dperm_right = _calc_k(calc_k_left_perm, right_perm=calc_k_right_perm)
        if perm_k <= k_optimal:
            if bilp_test: # extra bilp test
                kbilp_res = bilp.bilp(Dperm_right,method=bilp_method)
                kbilp = kbilp_res[0]
                if kbilp+perm_k <= k_optimal: # still need to continue
                    return True, 0, left_perm, add_left_perm, calc_k_left_perm, calc_k_right_perm
            else:
                return True, 0, left_perm, add_left_perm, calc_k_left_perm, calc_k_right_perm
            
        # pruned
        if prune_history:
            skipped = math.factorial(len(calc_k_right_perm))
        else:
            skipped = 0
        return False, skipped, left_perm, add_left_perm, calc_k_left_perm, calc_k_right_perm
    
    def _get_groups(D):
        def _no_left_calc_k(add_left_perms):
            chunk_results = [_calc_k(list(add_left_perm)) for add_left_perm in add_left_perms]
            i = 0
            add_left_perms_grouped = {}
            add_left_perms_grouped_best_ks = {}
            for kvalue,Dperm_right in chunk_results:
                sorted_items = sorted(add_left_perms[i])
                key = tuple(sorted_items)
                # initialize
                if key not in add_left_perms_grouped_best_ks:
                    add_left_perms_grouped_best_ks[key] = kvalue
                    add_left_perms_grouped[key] = []
                # re-initialize if we found a smaller k
                elif kvalue < add_left_perms_grouped_best_ks[key]:
                    add_left_perms_grouped_best_ks[key] = kvalue
                    add_left_perms_grouped[key] = []
                    
                # one of the best we've so far, so add to the group
                if kvalue == add_left_perms_grouped_best_ks[key]:
                    add_left_perms_grouped[key].append(add_left_perms[i])
                i+=1
            return add_left_perms_grouped, add_left_perms_grouped_best_ks
        
        add_left_perms = list(itertools.permutations(range(D.shape[0]), o))
        num_add_left_perms = len(add_left_perms)
        desired_number_chunks = 16
        chunk_length = round(num_add_left_perms*1./desired_number_chunks)
        add_left_perms_chunks = common.chunks(add_left_perms,chunk_length)
        futures = client.map(_no_left_calc_k,add_left_perms_chunks)
        results = client.gather(futures)
        add_left_perms_grouped = {}
        add_left_perms_grouped_best_ks = {}
        for chunk_add_left_perms_grouped, chunk_add_left_perms_grouped_best_ks in results:
            for key in chunk_add_left_perms_grouped.keys():
                if key not in add_left_perms_grouped or chunk_add_left_perms_grouped_best_ks[key] < add_left_perms_grouped_best_ks[key]:
                    add_left_perms_grouped[key] = chunk_add_left_perms_grouped[key]
                    add_left_perms_grouped_best_ks[key] = chunk_add_left_perms_grouped_best_ks[key]
                elif chunk_add_left_perms_grouped_best_ks[key] == add_left_perms_grouped_best_ks[key]:
                    add_left_perms_grouped[key].extend(chunk_add_left_perms_grouped[key])
        return add_left_perms_grouped, add_left_perms_grouped_best_ks
    
    def _multi_check(add_left_perms,left_perm):
        results = []
        add_left_perms_grouped = joblib.load(add_left_perms_grouped_file)
        for add_left_perm in add_left_perms:
            results.append(_check(left_perm,add_left_perm,add_left_perms_grouped))
        return results
    
    def _get_futures(left_perm,right_perm):
        add_left_perms = right_perm # list(itertools.permutations(right_perm, D.shape[0]/o))
        num_add_left_perms = len(right_perm) #common.nPr(len(right_perm),o)
        desired_number_chunks = 1000
        chunk_length = round(num_add_left_perms*1./desired_number_chunks)
        if chunk_length == 0:
            chunk_length = num_add_left_perms
        #chunk_length = 4000
        add_left_perms_chunks = common.chunks(add_left_perms,chunk_length)
        futures = client.map(_multi_check,
                             add_left_perms_chunks,
                             itertools.repeat(left_perm,len(add_left_perms_chunks)))
        return futures
    
    def _iterative_find_P():
        D = load_D()
        
        #try:
        #add_left_perms_grouped = joblib.load("/dev/shm/add_left_perms_grouped.joblib")
        #add_left_perms_grouped_best_ks = joblib.load("/dev/shm/add_left_perms_grouped_best_ks.joblib")
        #import pdb; pdb.set_trace()
        #for key in add_left_perms_grouped.keys():
        #    if len(add_left_perms_grouped[key]) == 0:
        #        import pdb; pdb.set_trace();
        #import pdb; pdb.set_trace()
        #except:
        add_left_perms_grouped, add_left_perms_grouped_best_ks = _get_groups(D)
        print('1')
        joblib.dump(add_left_perms_grouped,add_left_perms_grouped_file)
        joblib.dump(add_left_perms_grouped_best_ks,add_left_perms_grouped_best_ks_file)
        print(2)
        
        right_perm = [(key) for key in add_left_perms_grouped.keys()]
        print(3)
        P = []
        futures = _get_futures([],right_perm)
        print(4)
        seq = as_completed(futures)
        num_completed = 0
        for future in seq:
            if num_completed >= max_search_space:
                break
            if verbose:
                print("Futures remaining:",seq.count())
                print("Futures completed:",num_completed)
            j = 0
            try:
                for result_check, result_skipped, result_left_perm, result_add_left_perm, result_calc_k_left_perm, result_calc_k_right_perm in future.result():
                    num_completed += 1
                    if result_check:
                        if len(result_calc_k_right_perm) <= o or math.factorial(len(result_calc_k_right_perm)) <= t:
                            k, Pnew, search_perms = _exhaustive_search(result_calc_k_left_perm, result_calc_k_right_perm, [])
                            z = 0
                            result_right_perm = generate_right_perm(result_calc_k_left_perm)
                            for i, search_perm in enumerate(search_perms):
                                perm = result_left_perm + result_add_left_perm + result_right_perm + search_perm
                                # need to update Pnew at the right location
                                if k == k_optimal:
                                    P.append(perm)
                                elif k < k_optimal:
                                    raise Exception(
                                        "Error! The k value that you found was less than the optimal, which should never happen.")
                        else:
                            result_right_perm = generate_right_perm(result_calc_k_left_perm)
                            new_futures = _get_futures(result_left_perm + result_add_left_perm,result_right_perm)
                            for i, new_future in enumerate(new_futures):
                                seq.add(new_future)
                    elif prune_history:
                        if verbose:
                            print('pruned:',result_skipped)
                    j += 1
            except:
                client.recreate_error_locally(future)
               
        Pexpanded = []
        for perm in P:
            new_perms = [[]]
            for i,item in enumerate(perm):
                if type(item) == tuple:
                    new_new_perms = []
                    for perm_item in add_left_perms_grouped[item]:
                        if type(perm_item) == list:
                            perm_item = perm_item[0]
                        for old_perm in new_perms:                                
                            old_perm = copy.copy(old_perm)
                            old_perm.extend(perm_item)
                            new_new_perms.append(old_perm)
                    new_perms = new_new_perms
                else:
                    break
            # handle remainder
            item = perm[i:]
            if len(item) > 0:
                new_new_perms = []
                for old_perm in new_perms:
                    old_perm = copy.copy(old_perm)
                    old_perm.extend(item)
                    new_new_perms.append(old_perm)
                new_perms = new_new_perms
            Pexpanded.extend(new_perms)

        return Pexpanded
    
    P = _iterative_find_P()
    # cleanup
    try:
        os.remove(add_left_perms_grouped_best_ks_file)
    except:
        pass
    try:
        os.remove(add_left_perms_grouped_file)
    except:
        pass
    
    return k_optimal, P


import numpy as np
from numpy import ix_

from . import bilp
from . import common
from . import exact

import copy

# Generator allows us to perform local optimization in the context of a global solution
def generate_calc_k_spec(global_D,global_inxs):
    return lambda D,rem_inxs,partial_perm,prior_partial_perm: calc_k_local(D,rem_inxs,partial_perm,prior_partial_perm,global_D,global_inxs) 

def calc_k_local(D,rem_inxs,partial_perm,prior_partial_perm,global_D,global_inxs):
    rem_inxs_partial = tuple(rem_inxs[list(partial_perm)])
    rem_inxs_rest = np.delete(rem_inxs,partial_perm)
    Dspecial = common.permute_D(D,prior_partial_perm+rem_inxs_partial+tuple(rem_inxs_rest))
    # need to make the remaining end of matrix perfect
    # this currently is done with a loop, but needs to be done with numpy arrays
    # it also could be calculated with bilp for a tigther bound?
    for i1 in range(len(prior_partial_perm)+len(rem_inxs_partial),D.shape[0]):
        for j1 in range(len(prior_partial_perm)+len(rem_inxs_partial),D.shape[0]):
            if i1 == j1:
                Dspecial[i1,j1] = 0
            elif i1 > j1:
                Dspecial[i1,j1] = 0
            else:
                Dspecial[i1,j1] = 1

    global_D_copy = copy.copy(global_D)
    global_D_copy[ix_( global_inxs, global_inxs )] = Dspecial
    #return common.calc_k(global_D_copy),rem_inxs_partial,rem_inxs_rest
    return common.calc_k(Dspecial),rem_inxs_partial,rem_inxs_rest

def multi_compute_func(D, final_k, leave_out,target_search_space_at_leaf, entries, n, prior_partial_perm,rem_inxs,calc_k_spec):
    return [compute_func(D,final_k,leave_out,target_search_space_at_leaf,n,prior_partial_perm,partial_perm,rem_inxs,calc_k_spec) for partial_perm in entries]

def compute_func(D,final_k,leave_out,target_search_space_at_leaf,n,prior_partial_perm,partial_perm,rem_inxs,calc_k_spec):
    # local versions of variables
    res_k = None
    res_P = []
    skipped = 0
    searched = 0
    leaf = False

    sol_k,rem_inxs_partial,rem_inxs_rest = calc_k_spec(D,rem_inxs,partial_perm,prior_partial_perm)
    if sol_k <= final_k:
        rem_sol_k, rem_sol_P, rem_skipped, rem_searched, rem_leaf, exact_solution_found = find_P(D,target_search_space_at_leaf,leave_out,final_k-sol_k,
                                                                                                 n-leave_out,
                                                                                                 prior_partial_perm+rem_inxs_partial,
                                                                                                 calc_k_spec)
        skipped += rem_skipped
        searched += rem_searched
        if rem_sol_k != None:
            res_P_j = []
            for i,rem_perm in enumerate(rem_sol_P):
                if rem_leaf: # if this was a leaf, then do the conversion
                    rem_perm = tuple(rem_inxs_rest[list(rem_perm)])
                res_P_j.append(rem_inxs_partial+rem_perm)
            res_P = res_P_j # res_P + res_P_j
            res_k = sol_k + rem_sol_k
    else:
        skipped_exhaustive_search = exact.ExhaustiveSearch(n-leave_out)
        skipped += skipped_exhaustive_search.search_space_size
        exact_solution_found = True
    return res_k, res_P, skipped, searched, leaf, exact_solution_found

def find_P(D,target_search_space_at_leaf,leave_out,k,n,prior_partial_perm,calc_k_spec,num_parallel=1,client=None): 
    leaf = False
    skipped = 0
    searched = 0
    exact_solution_found = True

    exhaustive_search = exact.ExhaustiveSearch(n)
    # Base case 1: not large enough to leave out the number specified in leave_out
    # Base case 2: the search space is less than the target search space for a leaf node
    if n <= leave_out or exhaustive_search.search_space_size <= target_search_space_at_leaf:
        leaf = True
        exhaustive_search.leave_out = 0
        exhaustive_search.prepare_iterators(print_summary=False)
        D_copy = copy.copy(D)
        D = np.delete(D,prior_partial_perm,axis=0)
        D = np.delete(D,prior_partial_perm,axis=1)
        # We can solve this with an exact solution
        if exhaustive_search.search_space_size <= target_search_space_at_leaf:
            searched += exhaustive_search.search_space_size
            exhaustive_search.find_P(D,4)
            perm = tuple(exhaustive_search.P[0]) # pick one optimal permutation
            rem_inxs = np.arange(len(perm))+len(prior_partial_perm)
            k_global,rem_inxs_partial,rem_inxs_rest = calc_k_spec(D_copy,rem_inxs,perm,prior_partial_perm)
            if len(prior_partial_perm) > 0:
                prior_k = common.calc_k(common.permute_D(D_copy,prior_partial_perm))
            else:
                prior_k = 0
            k_adjusted = exhaustive_search.k #k_global-prior_k
            if k_adjusted > k:
                return None, [], skipped, searched, leaf, exact_solution_found
            elif k_adjusted == k:
                return exhaustive_search.k, exhaustive_search.P, skipped, searched, leaf, exact_solution_found
            else:
                #import pdb; pdb.set_trace()
                raise Exception("ERROR: Found k less than lower bound. Found k: " + str(int(exhaustive_search.k)) + " Lower bound: "+str(int(k)))
        else:
            solution = bilp.bilp(D,max_solutions=target_search_space_at_leaf)
            bilp_k = solution[0]
            bilp_P = (bilp_P-1).aslist()
            exact_solution_found = False
            return bilp_k, bilp_P, skipped, searched, leaf, exact_solution_found

    # Do the recursive step
    exhaustive_search.leave_out = leave_out
    exhaustive_search.prepare_iterators(print_summary=False)
    res_P = []
    rem_inxs = np.delete(np.arange(D.shape[0]),prior_partial_perm)
    res_k = None

    if client == None: 
        lambda_multi_compute_func = lambda entries: multi_compute_func(
            D, k, leave_out,target_search_space_at_leaf, entries, n, prior_partial_perm,rem_inxs,calc_k_spec)
        results = lambda_multi_compute_func(exhaustive_search.entries)
    else: # Right now we are only parallezing the outermost loop
        def len_chunks(l,n):
            return len(range(0,len(l),n))
        def chunks(l, n):
            chunks_ = []
            for i in range(0, len(l), n):
                #yield l[i:i + n]
                chunks_.append(l[i:i + n])
            return chunks_
        desired_num_chunks = 100
        size_of_chunk = int(len(exhaustive_search.entries)*1./desired_num_chunks)
        #size_of_chunk = int(1.*len(exhaustive_search.entries)/(num_parallel*1000))
        print('A. Begin breaking up search space into chunks of size',size_of_chunk)
        #chunks_ = chunks(exhaustive_search.entries,int(1.*len(exhaustive_search.entries)/num_parallel))
        chunks_ = chunks(exhaustive_search.entries,size_of_chunk)
        len_of_chunks = len_chunks(exhaustive_search.entries,size_of_chunk)
        print('A. Done')
        print('B. Begin map')
        values = client.map(multi_compute_func,
                            list(itertools.repeat(D,len_of_chunks)), 
                            list(itertools.repeat(k,len_of_chunks)), 
                            list(itertools.repeat(leave_out,len_of_chunks)),
                            list(itertools.repeat(target_search_space_at_leaf,len_of_chunks)),
                            chunks_,
                            list(itertools.repeat(n,len_of_chunks)),
                            list(itertools.repeat(prior_partial_perm,len_of_chunks)),
                            list(itertools.repeat(rem_inxs,len_of_chunks)),
                            list(itertools.repeat(calc_k_spec,len_of_chunks))
                           )
        print('B. Done')
        print('C. Begin gather')
        gather_results = client.gather(values)
        print('C. Done')
        results = []
        for gresults in gather_results:
            results.extend(gresults)

    # Now combine the results
    for ind_results in results:
        res_k_j, res_P_j, skipped_j, searched_j, leaf_j, exact_solution_found_j = ind_results
        if exact_solution_found_j == False:
            exact_solution_found = False
        if res_k_j != None:
            res_k = res_k_j
            res_P = res_P + res_P_j
        skipped += skipped_j
        searched += searched_j
    return res_k, res_P, skipped, searched, leaf, exact_solution_found


import networkx as nx
import numpy as np

def get_graph(P,names=None):
    mdg = nx.MultiDiGraph()
    
    dg = nx.DiGraph()
        
    #now add in the perm_i
    edges = {}
    for j in range(len(P)):
        perm_j = P[j,:]
        for i in range(len(perm_j)-1):
            i_t = perm_j[i]
            j_t = perm_j[i+1]
            if (i_t,j_t) not in edges:
                edges[(i_t,j_t)] = 0.0
            edges[(i_t,j_t)] += 1.0
            mdg.add_weighted_edges_from([(i_t,j_t, 1.0)],index=j)
            #colors.append("blue")
    for key in edges.keys():
        weight = edges[key]*1./len(P)
        i_t,j_t = key
        dg.add_weighted_edges_from([(i_t,j_t,weight)])

    #g.es["color"] = colors
    
    return mdg,dg
    
    


import itertools
import numpy as np
from numpy import ix_
from joblib import Parallel, delayed
import math
from joblib import parallel_backend
import tempfile

from . import bilp
from . import common

def reindex(orig_inxs, order_inxs):
    return list(np.array(orig_inxs)[order_inxs])

def find_k(D,bilp_method="mos2"):
    bilp_res = bilp.bilp(D,method=bilp_method)
    k_optimal = bilp_res[0]
    return k_optimal

def find_P(D, o, t, bilp_test=True, prune_history=None, check_and_recurse=True, bilp_method="mos2"):
    bilp_res = bilp.bilp(D,method=bilp_method)
    k_optimal = bilp_res[0]
    print("k",k_optimal)

    def _exhaustive_search(left_perm, search_perm, right_perm):
        all_search_perms = [list(perm) for perm in itertools.permutations(
            search_perm, len(search_perm))]

        def local_calc_k(local_search_perm): return _calc_k(
            left_perm + local_search_perm + right_perm)
        results = list(map(local_calc_k, all_search_perms))
        k_values = []
        for k_value, Dperm_right in results:
            k_values.append(k_value)
        k = np.min(k_values)
        P = [left_perm + opt_perm + right_perm for opt_perm in np.array(
            all_search_perms)[np.where(k_values == k)[0]].tolist()]
        return k, P

    def _calc_k(left_perm, right_perm=[]):
        perm = left_perm + right_perm
        Dperm = D[ix_(perm, perm)]
        # Left perm is optimal, but right perm is not, so we'll set it to the perfect submatrix.
        # This could be replaced to a call to bilp for a tighter bound
        if len(right_perm) > 0:
            rn = len(right_perm)
            right_inxs = np.arange(len(left_perm),Dperm.shape[0])
            Dperm_right = Dperm[ix_(right_inxs, right_inxs)]
            Dperm[ix_(right_inxs, right_inxs)] = np.triu(np.ones((rn, rn)), 1)
        else:
            Dperm_right = None
        Dperfect = np.triu(np.ones((D.shape[0], D.shape[0])), 1)
        k = int(np.sum(np.abs(Dperfect-Dperm)))
        return k, Dperm_right

    def _check_and_recurse(left_perm, add_left_perm, prev_right_perm):
        if type(add_left_perm) == tuple:
            add_left_perm = list(add_left_perm)
        
        # move the contents of add_left_perm out of prev_right_perm, this wrinkle is done to
        # save memory
        right_perm = list(set(prev_right_perm) - set(add_left_perm))
        
        perm_k, Dperm_right = _calc_k(left_perm+add_left_perm, right_perm=right_perm)
        if perm_k <= k_optimal:
            if bilp_test: # extra bilp test
                kbilp_res = bilp.bilp(Dperm_right,method=bilp_method)
                kbilp = kbilp_res[0]
                if kbilp+perm_k <= k_optimal: # still need to continue
                    return _find_P(left_perm+add_left_perm, right_perm)
            else:
                return _find_P(left_perm+add_left_perm, right_perm)
            
        # pruned
        skipped = math.factorial(len(right_perm))
        #if prune_history:
        #    skipped_perm = left_perm + add_left_perm
        #    print("prune",skipped_perm)
        return [], skipped
    
    def _check(left_perm,add_left_perm,prev_right_perm):
        if type(add_left_perm) == tuple:
            add_left_perm = list(add_left_perm)
        
        # move the contents of add_left_perm out of prev_right_perm, this wrinkle is done to
        # save memory
        right_perm = list(set(prev_right_perm) - set(add_left_perm))
        
        perm_k, Dperm_right = _calc_k(left_perm+add_left_perm, right_perm=right_perm)
        if perm_k <= k_optimal:
            if bilp_test: # extra bilp test
                kbilp_res = bilp.bilp(Dperm_right,method=bilp_method)
                kbilp = kbilp_res[0]
                if kbilp+perm_k <= k_optimal: # still need to continue
                    return True, 0, add_left_perm, right_perm
            else:
                return True, 0, add_left_perm, right_perm
            
        # pruned
        skipped = math.factorial(len(right_perm))
        #if prune_history:
        #    skipped_perm = left_perm + add_left_perm
        #    print("prune",skipped_perm)
        return False, skipped, add_left_perm, right_perm

    def _find_P(left_perm, right_perm):
        if len(right_perm) <= o or math.factorial(len(right_perm)) <= t:
            k, P = _exhaustive_search(left_perm, right_perm, [])
            # This could be collapsed, but I like to include it as a check of assumptions
            if k > k_optimal:
                return [], 0
            elif k == k_optimal:
                return P, 0
            else:
                raise Exception(
                    "Error! The k value that you found was less than the optimal, which should never happen.")

        # Recursive step
        skipped = 0
        if check_and_recurse: # Depth
            add_left_perms = [
                list(perm) for perm in itertools.permutations(list(right_perm), o)]
            right_perms = [list(set(right_perm) - set(add_left_perm))
                           for add_left_perm in add_left_perms]
            results = Parallel()(delayed(_check_and_recurse)(left_perm, add_left_perm, right_perm) for add_left_perm in add_left_perms)
        else: # Breadth
            add_left_perms = itertools.permutations(right_perm, o)
            num_add_left_perms = common.nPr(len(right_perm),o)
            results = Parallel()(delayed(_check_and_recurse)(left_perm, add_left_perm, right_perm) for add_left_perm in add_left_perms)
            results = map(_check,itertools.repeat(left_perm, num_add_left_perms), add_left_perms, itertools.repeat(right_perm, num_add_left_perms))
            not_skipped_add_left_perms = []
            not_skipped_right_perm = []
            i = 0                
            if prune_history:
                summary = {}
            for result_check,result_skipped, result_add_left_perm, result_right_perm in results:
                skipped += result_skipped
                if result_check:
                    not_skipped_add_left_perms.append(result_add_left_perm)
                    not_skipped_right_perm.append(result_right_perm)
                    # fill in a summary of what was not pruned
                    if prune_history:
                        for item in result_add_left_perm:
                            if item not in summary:
                                summary[item] = 0
                            summary[item] += 1
                i+=1
            if prune_history:
                print(summary)
                for item in right_perm:
                    if item not in summary:
                        print("prune",left_perm, -item, right_perm)
            results = Parallel()(delayed(_find_P)
                                 (left_perm + not_skipped_add_left_perms[i], not_skipped_right_perm[i]) for i in range(len(not_skipped_add_left_perms)))
        P = []
        for result_P, result_skipped in results:
            skipped += result_skipped
            P.extend(result_P)

        return P, skipped
    
    with parallel_backend("loky",n_jobs=14):
        P, skipped = _find_P([], list(range(D.shape[0])))
        return k_optimal, P, skipped



import itertools
import numpy as np
from numpy import ix_
import math
import tempfile
from dask import delayed
from dask.distributed import as_completed
from sklearn.externals import joblib
import copy
import uuid
import os
import math

from operator import itemgetter

from . import bilp
from . import common

import functools
print = functools.partial(print, flush=True)

def reindex(orig_inxs, order_inxs):
    return list(np.array(orig_inxs)[order_inxs])

def find_P(path_D, o, t, bilp_test=False, bilp_method="mos2", client=None, max_search_space=np.Inf, prune_history=False, uuid1=common.random_generator(),expand=True,verbose=False):
    def load_D():
        return np.genfromtxt(path_D, delimiter=",")
        
    orig_right_perm = list(range(load_D().shape[0]))
    bilp_res = bilp.bilp(load_D(),method=bilp_method)
    k_optimal = bilp_res[0]
    if verbose:
        print("k",k_optimal)
    #n = len(orig_right_perm)
    a = 0.5
    b = -1.0
    c = -k_optimal
    d = (b**2) - (4*a*c)
    n_worst = (-b+math.sqrt(d))/(2*a)
    o = min([round(n_worst),len(orig_right_perm)])
    o = 4
    if verbose:
        print('setting o to',o)

    add_left_perms_grouped_file = "/dev/shm/add_left_perms_grouped"+uuid1+".joblib"
    if verbose:
        print(add_left_perms_grouped_file)
    
    def _no_left_calc_k(add_left_perms):
        D = load_D()
        chunk_results = []
        #i = 0
        for add_left_perm in add_left_perms:
            chunk_results.append((add_left_perm,bilp.bilp(D[ix_(add_left_perm, add_left_perm)],method=bilp_method,max_solutions=1)))
            #if i % 10 == 0:
            #    print(i)
            #i += 1
            
        add_left_perms_grouped = []
        for add_left_perm,bilp_results in chunk_results:
            kvalue,P = bilp_results
            if kvalue > k_optimal:
                continue
            order_ixs = np.array(P[0])-1
            key = tuple(np.array(add_left_perm)[order_ixs].tolist())
            add_left_perms_grouped.append(key)
        return add_left_perms_grouped
    
    def _exhaustive_search(left_perm, search_perm, right_perm):
        all_search_perms = [list(perm) for perm in itertools.permutations(
            search_perm, len(search_perm))]

        def local_calc_k(local_search_perm): return _calc_k(
            left_perm + local_search_perm + right_perm)
        results = list(map(local_calc_k, all_search_perms))
        k_values = []
        for k_value, Dperm_right in results:
            k_values.append(k_value)
        k = np.min(k_values)
        optimal_inxs = np.where(k_values == k)[0]
        P = [left_perm + opt_perm + right_perm for opt_perm in np.array(
            all_search_perms)[optimal_inxs].tolist()]
        return k, P, [list(search_perm) for search_perm in np.array(all_search_perms)[optimal_inxs]]

    def _calc_k(left_perm, right_perm=[]):
        perm = left_perm + right_perm
        D = load_D()
        Dperm = D[ix_(perm, perm)]
        # Left perm is optimal, but right perm is not, so we'll set it to the perfect submatrix.
        # This could be replaced to a call to bilp for a tighter bound
        if len(right_perm) > 0:
            rn = len(right_perm)
            right_inxs = np.arange(len(left_perm),Dperm.shape[0])
            Dperm_right = Dperm[ix_(right_inxs, right_inxs)]
            Dperm[ix_(right_inxs, right_inxs)] = np.triu(np.ones((rn, rn)), 1)
        else:
            Dperm_right = None
        Dperfect = np.triu(np.ones((Dperm.shape[0], Dperm.shape[0])), 1)
        k = int(np.sum(np.abs(Dperfect-Dperm)))
        return k, Dperm_right
    
    def generate_right_perm(calc_k_left_perm):
        calc_k_right_perm = list(set(orig_right_perm) - set(calc_k_left_perm))
        _dict = {}
        for key in list(itertools.permutations(calc_k_right_perm, o)):
            _dict[tuple(sorted(key))] = True
        return list(_dict.keys())
    
    def generate_right_perm2(right_perm, result_calc_k_left_perm):
        next_right_perm = []
        result_calc_k_left_perm = set(result_calc_k_left_perm)
        for key in right_perm:
            if result_calc_k_left_perm.isdisjoint(key):
                next_right_perm.append(key)
        return next_right_perm
    
    def _check(left_perm,add_left_perm):
        def _convert(perm):
            actual_perm = []
            for key in perm:
                actual_perm += key
            return actual_perm  
    
        if type(add_left_perm) == tuple:
            add_left_perm = [add_left_perm]
        
        # move the contents of add_left_perm out of prev_right_perm, this wrinkle is done to
        # save memory
        if len(left_perm) == 0:
            calc_k_left_perm = _convert(add_left_perm)
        else:
            calc_k_left_perm = _convert(left_perm+add_left_perm)
        calc_k_right_perm = list(set(orig_right_perm) - set(calc_k_left_perm)) #list(set(prev_right_perm) - set(add_left_perm))
            
        perm_k, Dperm_right = _calc_k(calc_k_left_perm, right_perm=calc_k_right_perm)
        if perm_k <= k_optimal:
            if bilp_test: # extra bilp test
                kbilp_res = bilp.bilp(Dperm_right,method=bilp_method)
                kbilp = kbilp_res[0]
                if kbilp+perm_k <= k_optimal: # still need to continue
                    return True, 0, left_perm, add_left_perm, calc_k_left_perm, calc_k_right_perm
            else:
                return True, 0, left_perm, add_left_perm, calc_k_left_perm, calc_k_right_perm
            
        skipped = 0
        # pruned
        #if prune_history:
        #    skipped = math.factorial(len(calc_k_right_perm))
        #else:
        #    skipped = 0
        return False, skipped, left_perm, add_left_perm, calc_k_left_perm, calc_k_right_perm
    
    def _get_groups(D):
        add_left_perms = itertools.combinations(range(D.shape[0]), o)
        desired_number_chunks = 20
        num_add_left_perms = common.nCr(D.shape[0],o) #len(add_left_perms)
        chunk_length = round(num_add_left_perms*1./desired_number_chunks)
        add_left_perms_chunks = []
        for i in range(0, num_add_left_perms, chunk_length):
            add_left_perms_chunks.append(itertools.islice(add_left_perms,i,i+chunk_length)) #l[i:i + n])
        #add_left_perms_chunks = common.chunks(add_left_perms,chunk_length)
        if verbose:
            print('mapping',len(add_left_perms_chunks))
        #_no_left_calc_k(add_left_perms_chunks[0])
        futures = client.map(_no_left_calc_k,add_left_perms_chunks)
        if verbose:
            print('gathering')
        results = client.gather(futures)
        if verbose:
            print('done gathering')
        add_left_perms_grouped = []
        for chunk_add_left_perms_grouped in results:
            add_left_perms_grouped.extend(chunk_add_left_perms_grouped)
        """
        for chunk_add_left_perms_grouped, chunk_add_left_perms_grouped_best_ks in results:
            for key in chunk_add_left_perms_grouped.keys():
                if key not in add_left_perms_grouped or chunk_add_left_perms_grouped_best_ks[key] < add_left_perms_grouped_best_ks[key]:
                    add_left_perms_grouped[key] = chunk_add_left_perms_grouped[key]
                    add_left_perms_grouped_best_ks[key] = chunk_add_left_perms_grouped_best_ks[key]
                elif chunk_add_left_perms_grouped_best_ks[key] == add_left_perms_grouped_best_ks[key]:
                    add_left_perms_grouped[key].extend(chunk_add_left_perms_grouped[key])
        """
        return add_left_perms_grouped
    
    def _multi_check(add_left_perms,left_perm):
        results = []
        add_left_perms_grouped = joblib.load(add_left_perms_grouped_file)
        for add_left_perm in add_left_perms:
            results.append(_check(left_perm,add_left_perm))
        return results
    
    def _get_futures(left_perm,right_perm):
        add_left_perms = right_perm # list(itertools.permutations(right_perm, D.shape[0]/o))
        num_add_left_perms = len(right_perm) #common.nPr(len(right_perm),o)
        desired_number_chunks = 1000
        chunk_length = round(num_add_left_perms*1./desired_number_chunks)
        if chunk_length == 0:
            chunk_length = num_add_left_perms
        #chunk_length = 4000
        add_left_perms_chunks = common.chunks(add_left_perms,chunk_length)
        #import pdb; pdb.set_trace()
        futures = client.map(_multi_check,
                             add_left_perms_chunks,
                             itertools.repeat(left_perm,len(add_left_perms_chunks)))
        return futures
    
    def _iterative_find_P():
        D = load_D()
        
        #try:
        #add_left_perms_grouped = joblib.load("/dev/shm/add_left_perms_grouped.joblib")
        #add_left_perms_grouped_best_ks = joblib.load("/dev/shm/add_left_perms_grouped_best_ks.joblib")
        #import pdb; pdb.set_trace()
        #for key in add_left_perms_grouped.keys():
        #    if len(add_left_perms_grouped[key]) == 0:
        #        import pdb; pdb.set_trace();
        #import pdb; pdb.set_trace()
        #except:
        add_left_perms_grouped = _get_groups(D)
        joblib.dump(add_left_perms_grouped,add_left_perms_grouped_file)
        
        right_perm = add_left_perms_grouped # [(key) for key in add_left_perms_grouped.keys()]
        P = []
        futures = _get_futures([],right_perm)
        seq = as_completed(futures)
        num_completed = 0
        for future in seq:
            if num_completed >= max_search_space:
                print('Reached maximum search space. Stopping.')
                break
            if verbose:
                print("Futures remaining:",seq.count())
                print("Futures completed:",num_completed)
            j = 0
            try:
                for result_check, result_skipped, result_left_perm, result_add_left_perm, result_calc_k_left_perm, result_calc_k_right_perm in future.result():
                    num_completed += 1
                    if result_check:
                        if len(result_calc_k_right_perm) < o or math.factorial(len(result_calc_k_right_perm)) <= t:
                            k, Pnew, search_perms = _exhaustive_search(result_calc_k_left_perm, result_calc_k_right_perm, [])
                            z = 0
                            result_right_perm = generate_right_perm2(right_perm, result_calc_k_left_perm)
                            for i, search_perm in enumerate(search_perms):
                                perm = result_left_perm + result_add_left_perm + search_perm
                                # need to update Pnew at the right location
                                if k == k_optimal:
                                    P.append(perm)
                                elif k < k_optimal:
                                    raise Exception(
                                        "Error! The k value that you found was less than the optimal, which should never happen.")
                        else:
                            result_right_perm = generate_right_perm2(right_perm, result_calc_k_left_perm)
                            #(result_calc_k_left_perm)
                            new_futures = _get_futures(result_left_perm + result_add_left_perm,result_right_perm)
                            for i, new_future in enumerate(new_futures):
                                seq.add(new_future)
                    elif prune_history:
                        if verbose:
                            print('pruned prefix:',result_left_perm+result_add_left_perm)
                    j += 1
            except:
                client.recreate_error_locally(future)
               
        def _find_numbers(perms):
            solution_history = {}
            numbers = []
            Pexpanded = []
            for perm in perms:
                new_perms = [[]]
                numbers.append([])
                for i,item in enumerate(perm):
                    if type(item) == tuple:
                        new_new_perms = []
                        if item not in solution_history:
                            k, Pnew, search_perms = _exhaustive_search([], item, [])
                            solution_history[item] = Pnew
                        else:
                            Pnew = solution_history[item]
                        numbers[-1].append(len(Pnew))
                        if expand:            
                            for perm_item in Pnew:
                                #if type(perm_item) == list:
                                #    perm_item = perm_item[0]
                                for old_perm in new_perms:                                
                                    old_perm = copy.copy(old_perm)
                                    old_perm.extend(perm_item)
                                    new_new_perms.append(old_perm)
                            new_perms = new_new_perms
                    else:
                        break
                # handle remainder
                item = perm[i:]
                if len(item) > 0:
                    numbers[-1].append(1)
                    if expand:            
                        new_new_perms = []
                        for old_perm in new_perms:
                            old_perm = copy.copy(old_perm)
                            old_perm.extend(item)
                            new_new_perms.append(old_perm)
                        new_perms = new_new_perms
                Pexpanded.extend(new_perms)
            return numbers,Pexpanded
            
        desired_number_chunks = 12
        chunk_length = round(len(P)*1./desired_number_chunks)
        if chunk_length == 0:
            chunk_length = len(P)
        perms_chunks = common.chunks(P,chunk_length)
        if verbose:
            print('Finding final numbers')
        futures = client.map(_find_numbers,perms_chunks)
        results = client.gather(futures)
        numbers = []
        Pexpanded = []
        for chunk_numbers,chunk_Pexpanded in results:
            numbers.extend(chunk_numbers)
            Pexpanded.extend(chunk_Pexpanded)
        final_numbers = []
        p = 0
        for row in numbers:
            final_numbers.append(np.prod(row))
        p = np.sum(final_numbers)
        return p, Pexpanded
    
    p, P = _iterative_find_P()
    
    return k_optimal, p, P



import setuptools

setuptools.setup(
    name="pyrankability",
    version="0.0.10",
    author="Paul Anderson",
    author_email="pauleanderson@gmail.com",
    description="Rankability Toolbox",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/pypa/pyrankability",
    packages=setuptools.find_packages(),
    install_requires=[ "networkx", "numpy", "pandas" , "scipy", "dask", "toolz", "joblib", "distributed", "sklearn"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import itertools
import numpy as np
import pandas as pd
import copy

from . import common
from . import bilp
from . import pruning

def bilp_compute(D,inxs,max_solutions):
    return bilp.bilp(common.permute_D(D,inxs),max_solutions=max_solutions,method="mos2")

def find_P(D, max_solutions=1000, bilp_method="mos2"):
    k,P = bilp.bilp(D,method=bilp_method,max_solutions=max_solutions)
    return k, P

class LargeProblemSolver(common.Search):
    def __init__(self,D,num_iterations=10,max_num_starting_solutions=10,client=None,print_flag=True):
        self.num_iterations = num_iterations
        self.max_num_starting_solutions = max_num_starting_solutions
        self.D = D
        
        permutation_inxs = [np.random.permutation(D.shape[0]) for i in range(num_iterations)]
        if client != None:
            values = client.map(bilp_compute,
                                list(itertools.repeat(D,num_iterations)), 
                                permutation_inxs,
                                list(itertools.repeat(max_num_starting_solutions,num_iterations))
                               )
            results = client.gather(values)
        else:
            results = list(map(bilp_compute,
                          list(itertools.repeat(D,num_iterations)), 
                          permutation_inxs,
                          list(itertools.repeat(max_num_starting_solutions,num_iterations))
                         ))
        self.starting_results = results
        self.print_flag = print_flag
        # Now setup the P_to_search
        k = None
        P = []
        c = 0
        for ki,Pi in results:
            if k == None:
                k = ki
            if k != ki:
                raise "Error: k != ki"
            Pi_list = permutation_inxs[c][np.array(Pi)-1].tolist()
            P.extend(Pi_list)
            c+=1
        self.k = k
        self.P = []
        num_new_solutions = self.add_P(P)
        if self.print_flag:
            print("Number of new solutions added:",num_new_solutions)
        self.P_to_search = copy.deepcopy(self.P)
        self.record = {}
        self.P_searched = {}
        
    def add_P(self,P,check_k=False):
        if check_k:
            for Psolution in P:
                sol_k = common.calc_k(common.permute_D(self.D,Psolution))
                if sol_k != self.k:
                    raise Exception("k value is not correct when trying to add solution")
                    
        orig_num_solutions = len(self.P)
        self.P = pd.DataFrame(self.P+P).drop_duplicates().values.tolist()
        new_num_solutions = len(self.P)
        return new_num_solutions - orig_num_solutions
        
    def check_record(self,Psolution,global_inxs):
        key = (tuple(Psolution),tuple(global_inxs))
        if key in self.record:
            return True
        else:
            return False
    
    def add_record(self,Psolution,global_inxs,new_solutions):
        key = (tuple(Psolution),tuple(global_inxs))
        self.record[key] = new_solutions
        
    def find_P_pop(self,Psolution,size_of_chunk=8,client=None):
        num_new_solutions_added = 0
        if size_of_chunk > self.D.shape[0]:
            size_of_chunk = self.D.shape[0]
        chunks = common.chunks(np.arange(self.D.shape[0]),size_of_chunk)
        chunks_to_search = []
        for global_inxs in chunks:
            if self.check_record(Psolution,global_inxs) == False: # Not done yet
                chunks_to_search.append(global_inxs)
                
        if client == None:
            new_solutions_per_chunk = list(map(expand_solution,
                                               itertools.repeat(self.D,len(chunks_to_search)),
                                               itertools.repeat(Psolution,len(chunks_to_search)),
                                               chunks_to_search))
            for i,new_solutions_tuple in enumerate(new_solutions_per_chunk):
                k,new_solutions,skipped,searched,leaf,exact_solution_found = new_solutions_tuple
                global_inxs = chunks_to_search[i]
                self.add_record(Psolution,global_inxs,new_solutions)
                num_added = self.add_P(new_solutions,check_k=True)
                if self.print_flag:
                    print("Number of new solutions added:",num_added)
                    print("Number of skipped solutions:",skipped)
                    print("Number of searched solutions:",searched)
                num_new_solutions_added += num_added
        
        if tuple(Psolution) not in self.P_searched:
            self.P_searched[tuple(Psolution)] = 0
        self.P_searched[tuple(Psolution)] += 1
                
        return num_new_solutions_added
    
    def find_P(self,size_of_chunk=8,client=None):
        while len(self.P_to_search) > 0:
            Psolution = np.array(self.P_to_search.pop())
            if tuple(Psolution) not in self.P_searched:
                if self.print_flag:
                    print("Performing local search for",tuple(Psolution))
                num_new_solutions = self.find_P_pop(Psolution,size_of_chunk=size_of_chunk,client=client)
                #if self.print_flag:
                #    print("Total number of new solutions:",num_new_solutions)
                
    def other_to_dict(self):
        return {"exact_solution_found": False}

def expand_solution(D,Psolution,global_inxs,target_search_space_at_leaf=1000,leave_out=4,client=None):
    subset = Psolution[global_inxs]
    Dspec = common.permute_D(D,subset)
    k_spec = common.calc_k(Dspec)
    global_D = common.permute_D(D,Psolution)
    global_sol_k = common.calc_k(global_D)
    
    calc_k_spec = pruning.generate_calc_k_spec(global_D,global_inxs)
    k, P, skipped, searched, leaf, exact_solution_found = pruning.find_P(
        Dspec,target_search_space_at_leaf,leave_out,k_spec,Dspec.shape[0],(),calc_k_spec,client=client)
    new_solutions = []
    for local_solution in P:
        new_solution = copy.copy(Psolution)
        new_solution[global_inxs] = subset[np.array(local_solution)]
        Dperm = common.permute_D(D,new_solution)
        new_solutions.append(list(new_solution))
    return k,new_solutions,skipped,searched,leaf,exact_solution_found





from . import analysis
from . import approximate
from . import bilp
from . import lp
from . import common
from . import exact
from . import pruning
from . import pruning_paper
from . import pruning_paper_dask
from . import pruning_paper_dask2
from . import pruning_paper_dask3


import itertools
import numpy as np
from scipy.misc import comb
import math

from dask.distributed import Client

from . import common
from . import bilp
from . import pruning

#n=5
#k=15

def find_P_simple_parallel(D):
    if D.shape[0] > 10:
        raise "D cannot have more than 10 items for this method"
    search = pyrankability.exact.ExhaustiveSearch(D.shape[0])
    search.leave_out = 4
    search.prepare_iterators()
    search.find_P(D,2)
    
    return search.k,search.P

def find_P_simple(D):
    # TODO: need to implement a simple loop
    pass

class ExhaustiveSearch(common.Search):
    def __init__(self,D,dask_connection_string="127.0.0.1:8786",leave_out = 4):
        self.D = D
        n = D.shape[0]
        self.n = n
        self.dask_connection_string = dask_connection_string
        self.search_space_size = math.factorial(n)
        self.solution_size = n
        self.max_leave_out = self.solution_size - 1
        self.leave_out = leave_out
    
    def print_solution_space_summary(self,leave_out):
        print('Total search space of the problem:',self.search_space_size)
        print('Number of workers',self.calc_num_workers(leave_out))
        print('Work per worker',self.calc_work_per_worker(leave_out))
        #print('Number of subproblems/worker',total_size//work_per_worker)
        print('Achieved by leaving',leave_out,'items out')
        print('Check num_workers x work_per_worker equals total search space:',
             self.calc_num_workers(leave_out)*self.calc_work_per_worker(leave_out) == self.search_space_size)
    
    def calc_work_per_worker(self,leave_out):
        return math.factorial(self.n-leave_out)
    
    def calc_num_workers(self,leave_out):
        return nPr(self.n,leave_out) #math.factorial(leave_out)
    
    def determine_leave_out(self,num_parallel,min_work_per_worker):
        # Decide if we should devide the problem into subtasks and 
        # if so how many and with how much work
        single_thread = True
        leave_out = 1
        while leave_out < self.max_leave_out:
            work_per_worker = self.calc_work_per_worker(leave_out) #nCr(self.total_num_indices-leave_out,k)
            num_subtasks = self.calc_num_workers(leave_out) #nCr(self.solution_size,leave_out)
            if num_subtasks > num_parallel and work_per_worker > min_work_per_worker:
                single_thread = False # Don't run as a single thread
                break
            leave_out += 1
            
        return leave_out, single_thread
    
    def generate_src_inxs(self):
        return list(range(self.solution_size))
    
    def generate_leave_out_iter(self):
        return itertools.permutations(self.generate_src_inxs(),self.leave_out)
    
    def generate_iter(self,subproblem):
        return itertools.permutations(subproblem)
    
    def autoselect_leaveout(self,min_num_workers=2,min_work_per_worker=1):
        leave_out, single_thread = self.determine_leave_out(min_num_workers,min_work_per_worker)
        self.leave_out = leave_out
        self.single_thread = single_thread
    
    def prepare_iterators(self,print_summary=False):
        if print_summary:
            self.print_solution_space_summary(self.leave_out)
        
        # Now create the iterators that can run in parallel
        src_inxs = self.generate_src_inxs()
        leave_out_iter = self.generate_leave_out_iter()#itertools.permutations(src_inxs,leave_out)
        subproblems = []
        iters = []
        entries = []
        for entry in leave_out_iter:
            src_inxs_set = set(src_inxs)
            entry_set = set(entry)
            subproblems.append(src_inxs_set-entry_set)
            entries.append(entry)
            iters.append(self.generate_iter(subproblems[-1]))        
        
        self.solutions_iters = iters
        self.subproblems = subproblems
        self.entries = entries
    
    @staticmethod
    def find_P_parallel(D,solutions_iter,entry):
        k = np.Inf
        P = []
        for j,solution in enumerate(solutions_iter):
            perm = entry+solution
            
            sol_k = common.calc_k(common.permute_D(D,perm))
            if sol_k < k:
                k = sol_k
                P = []
            if sol_k == k:
                P.append(tuple(np.array(perm)))
        return int(k),P
               
    def get_find_P_parallel_func(self):
        return lambda solutions_iter,entry: ExhaustiveSearch.find_P_parallel(self.D,solutions_iter,entry)

    def find_P(self):
        """find_P finds the solutions to the rankability problem.
        """
        self.prepare_iterators()

        opt_P = []
        opt_k = np.Inf
        client = Client(self.dask_connection_string)
        futures = client.map(self.get_find_P_parallel_func(),self.solutions_iters,self.entries)
        results = client.gather(futures)
        for k,P in results:
            if k < opt_k:
                opt_P = []
                opt_k = k
            if k == opt_k:
                opt_P.extend(P)
        self.k = opt_k
        self.P = opt_P
        
class RecusiveKBoundedSearch(common.Search):
    def __init__(self,D,target_search_space_at_leaf=1000,leave_out=4):
        self.target_search_space_at_leaf = target_search_space_at_leaf
        self.leave_out = leave_out
        self.D = D
        bilp_res = bilp.bilp(self.D)
        self.k = bilp_res[0]
        
    def find_P(self,client=None):
        calc_k_spec = pruning.generate_calc_k_spec(self.D,np.arange(self.D.shape[0]))
        self.k,self.P,self.skipped,self.searched,leaf,self.exact_solution_found = pruning.find_P(
            self.D,self.target_search_space_at_leaf,self.leave_out,self.k,self.D.shape[0],(),calc_k_spec,num_parallel=18,client=client)
    